@@ -2,21 +2,21 @@
 // This software is released under the 2-Clause BSD license, included
 // below.
 //
-// Copyright (c) 2019, Aous Naman 
+// Copyright (c) 2019, Aous Naman
 // Copyright (c) 2019, Kakadu Software Pty Ltd, Australia
 // Copyright (c) 2019, The University of New South Wales, Australia
-// 
+//
 // Redistribution and use in source and binary forms, with or without
 // modification, are permitted provided that the following conditions are
 // met:
-// 
+//
 // 1. Redistributions of source code must retain the above copyright
 // notice, this list of conditions and the following disclaimer.
-// 
+//
 // 2. Redistributions in binary form must reproduce the above copyright
 // notice, this list of conditions and the following disclaimer in the
 // documentation and/or other materials provided with the distribution.
-// 
+//
 // THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 // IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 // TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
@@ -100,6 +100,7 @@ namespace ojph {
     OJPH_TILEPART_RESOLUTIONS  = 0x1,
     OJPH_TILEPART_COMPONENTS   = 0x2,
     OJPH_TILEPART_LAYERS       = 0x4, // these are meaningless with HTJ2K
+    OJPH_TILEPART_MASK         = 0x3, // mask used for testing
   };
 
   namespace local {
@@ -108,7 +109,6 @@ namespace ojph {
     struct param_siz;
     struct param_cod;
     struct param_qcd;
-    struct param_qcc;
     struct param_cap;
     struct param_sot;
     struct param_tlm;
@@ -174,9 +174,12 @@ namespace ojph {
       };
 
     public:
-      param_siz()
+      param_siz() { init(); }
+      ~param_siz() { destroy(); }
+
+      void init()
       {
-        Lsiz = Csiz = 0;        
+        Lsiz = Csiz = 0;
         Xsiz = Ysiz = XOsiz = YOsiz = XTsiz = YTsiz = XTOsiz = YTOsiz = 0;
         skipped_resolutions = 0;
         memset(store, 0, sizeof(store));
@@ -185,13 +188,11 @@ namespace ojph {
         dfs = NULL;
         Rsiz = RSIZ_HT_FLAG;
         cptr = store;
-        old_Csiz = 4;
+        old_Csiz = sizeof(store) / sizeof(siz_comp_info);
       }
 
-      ~param_siz()
-      {
-        if (cptr != store) delete[] cptr;
-      }
+      void destroy()
+      { if (cptr != store) { delete[] cptr; cptr = NULL; } }
 
       void set_num_components(ui32 num_comps)
       {
@@ -223,7 +224,7 @@ namespace ojph {
         if (XTsiz == 0 && YTsiz == 0)
         { XTsiz = Xsiz + XOsiz; YTsiz = Ysiz + YOsiz; }
         if (Xsiz == 0 || Ysiz == 0 || XTsiz == 0 || YTsiz == 0)
-          OJPH_ERROR(0x00040001, 
+          OJPH_ERROR(0x00040001,
             "You cannot set image extent nor tile size to zero");
         if (XTOsiz > XOsiz || YTOsiz > YOsiz)
           OJPH_ERROR(0x00040002,
@@ -261,7 +262,7 @@ namespace ojph {
 
       void set_skipped_resolutions(ui32 skipped_resolutions)
       { this->skipped_resolutions = skipped_resolutions; }
-      
+
       ui32 get_width(ui32 comp_num) const
       {
         assert(comp_num < get_num_components());
@@ -269,7 +270,7 @@ namespace ojph {
         ui32 t = ojph_div_ceil(Xsiz, ds) - ojph_div_ceil(XOsiz, ds);
         return t;
       }
-      
+
       ui32 get_height(ui32 comp_num) const
       {
         assert(comp_num < get_num_components());
@@ -328,6 +329,15 @@ namespace ojph {
     ///////////////////////////////////////////////////////////////////////////
     struct cod_SPcod
     {
+      cod_SPcod() {
+        num_decomp = 5;
+        block_width = 4;    // 64
+        block_height = 4;   // 64
+        block_style = 0x40; // HT mode
+        wavelet_trans = 0;  // reversible 5 / 3
+        memset(precinct_size, 0, sizeof(precinct_size));
+      }
+
       ui8 num_decomp;
       ui8 block_width;
       ui8 block_height;
@@ -337,7 +347,7 @@ namespace ojph {
 
       size get_log_block_dims() const
       { return size(block_width + 2, block_height + 2); }
-      size get_block_dims() const 
+      size get_block_dims() const
       { size t = get_log_block_dims(); return size(1 << t.w, 1 << t.h); }
       size get_log_precinct_size(ui32 res_num) const
       {
@@ -350,6 +360,7 @@ namespace ojph {
     ///////////////////////////////////////////////////////////////////////////
     struct cod_SGcod
     {
+      cod_SGcod() : prog_order(OJPH_PO_RPCL), num_layers(1), mc_trans(0) {}
       ui8 prog_order;
       ui16 num_layers;
       ui8 mc_trans;
@@ -359,8 +370,12 @@ namespace ojph {
     struct param_cod
     {
       // serves for both COD and COC markers
-
       friend ::ojph::param_cod;
+      enum default_comp_num : ui16 {
+        OJPH_COD_UNKNOWN = 65534,
+        OJPH_COD_DEFAULT = 65535
+      };
+
       ////////////////////////////////////////
       enum BLOCK_CODING_STYLES {
         VERT_CAUSAL_MODE = 0x8,
@@ -371,8 +386,8 @@ namespace ojph {
         UNDEFINED = 0,
         COD_MAIN  = 1,
         COC_MAIN  = 2,
-        COD_TILE  = 3,
-        COC_TILE  = 4
+        COD_TILE  = 3,  // not implemented
+        COC_TILE  = 4   // not implemented
       };
       ////////////////////////////////////////
       enum dwt_type : ui8 {
@@ -381,25 +396,24 @@ namespace ojph {
       };
 
     public: // COD_MAIN and COC_MAIN common functions
+      param_cod(param_cod* top_cod = NULL, ui16 comp_idx = OJPH_COD_DEFAULT)
+      { avail = NULL; init(top_cod, comp_idx); }
+      ~param_cod() { destroy(); }
+
       ////////////////////////////////////////
-      param_cod()
+      void restart()
       {
-        memset(this, 0, sizeof(param_cod));
-        SPcod.block_style = HT_MODE;
-        SGCod.prog_order = OJPH_PO_RPCL;
-        SGCod.num_layers = 1;
-        SGCod.mc_trans = 0;
-        SPcod.num_decomp = 5;
-        SPcod.block_width = 4; //64
-        SPcod.block_height = 4; //64
-        next = NULL;
+        param_cod** p = &avail; // move next to the end of avail
+        while (*p != NULL)
+          p = &((*p)->next);
+        *p = next;
+        this->init(top_cod, OJPH_COD_DEFAULT);
       }
 
       ////////////////////////////////////////
       void set_reversible(bool reversible)
       {
-        assert(type == UNDEFINED || type == COD_MAIN);
-        type = COD_MAIN;
+        assert(type == UNDEFINED || type == COD_MAIN || type == COC_MAIN);
         SPcod.wavelet_trans = reversible ? DWT_REV53 : DWT_IRV97;
       }
 
@@ -408,15 +422,13 @@ namespace ojph {
       {
         assert(val == 0 || val == 1);
         assert(type == UNDEFINED || type == COD_MAIN);
-        type = COD_MAIN;
         SGCod.mc_trans = val;
       }
 
       ////////////////////////////////////////
       void check_validity(const param_siz& siz)
       {
-        assert(type == UNDEFINED || type == COD_MAIN);
-        type = COD_MAIN;
+        assert(type == COD_MAIN);
 
         //check that colour transform and match number of components and
         // downsampling
@@ -428,21 +440,40 @@ namespace ojph {
 
         if (SGCod.mc_trans == 1)
         {
-          bool test = false;
+          bool test_signedness = false;
+          bool test_bit_depth = false;
+          bool test_downsampling = false;
           point p = siz.get_downsampling(0);
+          ui32 bit_depth = siz.get_bit_depth(0);
+          bool is_signed = siz.is_signed(0);
           for (ui32 i = 1; i < 3; ++i)
           {
             point p1 = siz.get_downsampling(i);
-            test = test || (p.x != p1.x || p.y != p1.y);
+            test_downsampling = test_downsampling
+              || (p.x != p1.x || p.y != p1.y);
+            test_bit_depth = test_bit_depth
+              || (bit_depth != siz.get_bit_depth(i));
+            test_signedness = test_signedness
+              || (is_signed != siz.is_signed(i));
           }
-          if (test)
+          if (test_downsampling)
             OJPH_ERROR(0x00040012,
-              "when color transform is used, the first 3 colour "
-              "components must have the same downsampling.");
+              "when color transform is used, the first 3 colour components "
+              "must have the same downsampling factor.");
+          if (test_bit_depth)
+            OJPH_ERROR(0x00040014,
+              "when color transform is used, the first 3 colour components "
+              "must have the same bit depth.");
+          if (test_signedness)
+            OJPH_ERROR(0x00040015,
+              "when color transform is used, the first 3 colour components "
+              "must have the same signedness (signed or unsigned).");
+
         }
 
         //check the progression order matches downsampling
-        if (SGCod.prog_order == 2 || SGCod.prog_order == 3)
+        if (SGCod.prog_order == OJPH_PO_RPCL ||
+            SGCod.prog_order == OJPH_PO_PCRL)
         {
           ui32 num_comps = siz.get_num_components();
           for (ui32 i = 0; i < num_comps; ++i)
@@ -463,7 +494,7 @@ namespace ojph {
         else if (type == COC_MAIN)
         {
           if (is_dfs_defined())
-            return parent->get_num_decompositions();
+            return top_cod->get_num_decompositions();
           else
             return SPcod.num_decomp;
         }
@@ -485,9 +516,17 @@ namespace ojph {
       ui8 get_wavelet_kern() const
       { return SPcod.wavelet_trans; }
 
+      ////////////////////////////////////////
+      bool is_reversible() const;
+
       ////////////////////////////////////////
       bool is_employing_color_transform() const
-      { return (SGCod.mc_trans == 1); }
+      {
+        if (type == COD_MAIN || type == COD_TILE)
+          return (SGCod.mc_trans == 1);
+        else
+          return top_cod->is_employing_color_transform();
+      }
 
       ////////////////////////////////////////
       size get_precinct_size(ui32 res_num) const
@@ -498,7 +537,7 @@ namespace ojph {
 
       ////////////////////////////////////////
       size get_log_precinct_size(ui32 res_num) const
-      { 
+      {
         if (Scod & 1)
           return SPcod.get_log_precinct_size(res_num);
         else
@@ -507,60 +546,54 @@ namespace ojph {
 
       ////////////////////////////////////////
       bool packets_may_use_sop() const
-      { 
-        if (parent)
-          return (parent->Scod & 2) == 2; 
-        else
+      {
+        if (type == COD_MAIN || type == COD_TILE)
           return (Scod & 2) == 2;
+        return false;
       }
 
       ////////////////////////////////////////
       bool packets_use_eph() const
-      { 
-        if (parent)
-          return (parent->Scod & 4) == 4;
-        else
+      {
+        if (type == COD_MAIN || type == COD_TILE)
           return (Scod & 4) == 4;
+        return false;
       }
 
       ////////////////////////////////////////
-      ui32 propose_precision(const param_siz* siz, ui32 comp_num) const;
+      bool get_block_vertical_causality() const
+      { return (SPcod.block_style & local::param_cod::VERT_CAUSAL_MODE) != 0; }
 
       ////////////////////////////////////////
       bool write(outfile_base *file);
 
       ////////////////////////////////////////
-      void read(infile_base *file, cod_type type);
+      bool write_coc(outfile_base *file, ui32 num_comps);
 
       ////////////////////////////////////////
-      void read(infile_base* file, cod_type type, ui32 num_comps, 
-                param_cod* cod);
+      void read(infile_base *file);
 
       ////////////////////////////////////////
-      void update_atk(const param_atk* atk);
+      void read_coc(infile_base* file, ui32 num_comps, param_cod* top_cod);
 
       ////////////////////////////////////////
-      void link_cod(const param_cod* coc)
-      { this->next = coc; }
+      void update_atk(param_atk* atk);
 
       ////////////////////////////////////////
-      const param_cod* get_cod(ui32 comp_num) const
-      {
-        const param_cod* result = this->next;
-        while (result != NULL && result->get_comp_num() != comp_num)
-          result = result->next;
-        if (result)
-          return result;
-        else
-          return this;
-      }
+      const param_cod* get_coc(ui32 comp_idx) const;
+
+      ////////////////////////////////////////
+      param_cod* get_coc(ui32 comp_idx);
+
+      ////////////////////////////////////////
+      param_cod* add_coc_object(ui32 comp_idx);
 
       ////////////////////////////////////////
       const param_atk* access_atk() const { return atk; }
 
     public: // COC_MAIN only functions
       ////////////////////////////////////////
-      bool is_dfs_defined() const 
+      bool is_dfs_defined() const
       { return (SPcod.num_decomp & 0x80) != 0; }
 
       ////////////////////////////////////////
@@ -568,21 +601,55 @@ namespace ojph {
       { return SPcod.num_decomp & 0xF; }
 
       ////////////////////////////////////////
-      ui32 get_comp_num() const
-      { assert(type == COC_MAIN); return comp_num; }
+      ui32 get_comp_idx() const
+      {
+        assert((type == COC_MAIN && comp_idx != OJPH_COD_DEFAULT) ||
+               (type == COD_MAIN && comp_idx == OJPH_COD_DEFAULT));
+        return comp_idx;
+      }
 
+    private:
+      ////////////////////////////////////////
+      void init(param_cod* top_cod, ui16 comp_idx)
+      {
+        type = top_cod ? COC_MAIN : COD_MAIN;
+        Lcod = 0;
+        Scod = 0;
+        next = NULL;
+        atk = NULL;
+        this->top_cod = top_cod;
+        this->comp_idx = comp_idx;
+      }
+
+      ////////////////////////////////////////
+      void destroy() {
+        if (avail)
+          delete avail;
+        if (next) {
+          delete next;
+          next = NULL;
+        }
+      }
+
+    private:
+      bool internal_write_coc(outfile_base *file, ui32 num_comps);
+
+    ////////////////////////////////////////
     private: // Common variables
       cod_type type;        // The type of this cod structure
       ui16 Lcod;            // serves as Lcod and Scod
       ui8 Scod;             // serves as Scod and Scoc
       cod_SGcod SGCod;      // Used in COD and copied to COC
       cod_SPcod SPcod;      // serves as SPcod and SPcoc
-      const param_cod* next;// to chain coc parameters to cod
+      param_cod* next;      // to chain coc parameters to cod
       const param_atk* atk; // used to read transform information
 
     private: // COC only variables
-      param_cod* parent;    // parent COD structure
-      ui16 comp_num;        // component index of this COC structure
+      param_cod* top_cod;   // parent COD structure
+      ui16 comp_idx;        // component index of this COC structure
+
+    private: // on restart, already allocated param_cod objs are stored here
+      param_cod* avail;
     };
 
     ///////////////////////////////////////////////////////////////////////////
@@ -594,96 +661,119 @@ namespace ojph {
     ///////////////////////////////////////////////////////////////////////////
     struct param_qcd
     {
+      // serves for both QCD and QCC markers
       friend ::ojph::param_qcd;
+      enum default_comp_num : ui16 {
+        OJPH_QCD_UNKNOWN = 65534,
+        OJPH_QCD_DEFAULT = 65535
+      };
+
+      ////////////////////////////////////////
+      enum qcd_type : ui8 {
+        UNDEFINED = 0,
+        QCD_MAIN  = 1,
+        QCC_MAIN  = 2,
+        QCD_TILE  = 3,  // not implemented
+        QCC_TILE  = 4   // not implemented
+      };
+
     public:
-      param_qcd()
-      { 
-        Lqcd = 0;
-        Sqcd = 0;
-        memset(u16_SPqcd, 0, sizeof(u16_SPqcd));
-        num_subbands = 0;
-        base_delta = -1.0f;
-      }
+      param_qcd(param_qcd* top_qcd = NULL, ui16 comp_idx = OJPH_QCD_DEFAULT)
+      { avail = NULL; init(top_qcd, comp_idx); }
+      ~param_qcd() { destroy(); }
 
-      void set_delta(float delta) { base_delta = delta; }
-
-      void check_validity(const param_siz& siz, const param_cod& cod)
+      ////////////////////////////////////////
+      void restart()
       {
-        ui32 num_decomps = cod.get_num_decompositions();
-        num_subbands = 1 + 3 * num_decomps;
-        if (cod.get_wavelet_kern() == param_cod::DWT_REV53)
-        {
-          ui32 bit_depth = 0;
-          for (ui32 i = 0; i < siz.get_num_components(); ++i)
-            bit_depth = ojph_max(bit_depth, siz.get_bit_depth(i));
-          set_rev_quant(num_decomps, bit_depth,
-            cod.is_employing_color_transform());
-        }
-        else if (cod.get_wavelet_kern() == param_cod::DWT_IRV97)
-        {
-          if (base_delta == -1.0f) {
-            ui32 bit_depth = 0;
-            for (ui32 i = 0; i < siz.get_num_components(); ++i)
-              bit_depth =
-              ojph_max(bit_depth, siz.get_bit_depth(i) + siz.is_signed(i));
-            base_delta = 1.0f / (float)(1 << bit_depth);
-          }
-          set_irrev_quant(num_decomps);
-        }
-        else
-          assert(0);
+        param_qcd** p = &avail; // move next to the end of avail
+        while (*p != NULL)
+          p = &((*p)->next);
+        *p = next;
+        this->init(top_qcd, OJPH_QCD_DEFAULT);
       }
+
+      void check_validity(const param_siz& siz, const param_cod& cod);
+      void set_delta(float delta) { base_delta = delta; }
+      void set_delta(ui32 comp_idx, float delta);
       ui32 get_num_guard_bits() const;
-      ui32 get_MAGBp() const;
+      ui32 get_MAGB() const;
       ui32 get_Kmax(const param_dfs* dfs, ui32 num_decompositions,
                     ui32 resolution, ui32 subband) const;
-      float irrev_get_delta(const param_dfs* dfs,
+      ui32 propose_precision(const param_cod* cod) const;
+      float get_irrev_delta(const param_dfs* dfs,
                             ui32 num_decompositions,
                             ui32 resolution, ui32 subband) const;
-
       bool write(outfile_base *file);
+      bool write_qcc(outfile_base *file, ui32 num_comps);
       void read(infile_base *file);
+      void read_qcc(infile_base *file, ui32 num_comps);
 
-    protected:
-      void set_rev_quant(ui32 num_decomps, ui32 bit_depth, 
+      param_qcd* get_qcc(ui32 comp_idx);
+      const param_qcd* get_qcc(ui32 comp_idx) const;
+      param_qcd* add_qcc_object(ui32 comp_idx);
+      ui16 get_comp_idx() const { return comp_idx; }
+
+    private:
+      ////////////////////////////////////////
+      void init(param_qcd* top_qcd, ui16 comp_idx)
+      {
+        type = top_qcd ? QCC_MAIN : QCD_MAIN;
+        Lqcd = 0;
+        Sqcd = 0;
+        memset(&SPqcd, 0, sizeof(SPqcd));
+        num_subbands = 0;
+        base_delta = -1.0f;
+        enabled = true;
+        next = NULL;
+        this->top_qcd = top_qcd;
+        this->comp_idx = comp_idx;
+      }
+
+      ////////////////////////////////////////
+      void destroy() {
+        if (avail)
+          delete avail;
+        if (next)
+        {
+          delete next;
+          next = NULL;
+        }
+      }
+
+    private:
+      void set_rev_quant(ui32 num_decomps, ui32 bit_depth,
                          bool is_employing_color_transform);
       void set_irrev_quant(ui32 num_decomps);
+      ui32 get_largest_Kmax() const;
+      bool internal_write_qcc(outfile_base *file, ui32 num_comps);
+      void trim_non_existing_components(ui32 num_comps);
 
       ui8 decode_SPqcd(ui8 v) const
       { return (ui8)(v >> 3); }
       ui8 encode_SPqcd(ui8 v) const
       { return (ui8)(v << 3); }
-   protected:
+
+    private: // QCD variables
+      qcd_type type;
       ui16 Lqcd;
       ui8 Sqcd;
       union
       {
-        ui8 u8_SPqcd[97];
-        ui16 u16_SPqcd[97];
-      };
-      ui32 num_subbands;        // number of subbands
-      float base_delta;         // base quantization step size -- all other
-                                // step sizes are derived from it.
-    };
+        ui8 u8[97];
+        ui16 u16[97];
+      } SPqcd;
+      ui32 num_subbands;  // number of subbands
+      float base_delta;   // base quantization step size -- all other
+                          // step sizes are derived from it.
+      bool enabled;       // enabled if two, and ignored if false
+      param_qcd *next;    // pointer to create chains of qcc marker segments
+      param_qcd *top_qcd; // pointer to the top QCD (this is the default)
 
-    ///////////////////////////////////////////////////////////////////////////
-    //
-    //
-    //
-    //
-    //
-    ///////////////////////////////////////////////////////////////////////////
-    struct param_qcc : public param_qcd
-    {
-    public:
-      param_qcc() : param_qcd()
-      { comp_idx = 0; }
+    private: // QCC only variables
+      ui16 comp_idx;
 
-      ui16 get_comp_num() { return comp_idx; }
-      void read(infile_base *file, ui32 num_comps);
-
-    protected:
-        ui16 comp_idx;
+    private:  // on restart, already allocated param_qcd objs are stored here
+      param_qcd* avail;
     };
 
     ///////////////////////////////////////////////////////////////////////////
@@ -697,33 +787,53 @@ namespace ojph {
     struct param_nlt
     {
       using special_comp_num = ojph::param_nlt::special_comp_num;
+      using nonlinearity = ojph::param_nlt::nonlinearity;
     public:
-      param_nlt() { 
-        Lnlt = 6;
-        Cnlt = special_comp_num::ALL_COMPS; // default
-        BDnlt = 0;
-        Tnlt = 3;
-        enabled = false; next = NULL; alloced_next = false;
-      }
+      param_nlt() { avail = NULL; init(); }
+      ~param_nlt() { destroy(); }
 
-      ~param_nlt() {
-        if (next && alloced_next) {
-          delete next;
-          alloced_next = false;
-          next = NULL;
-        }
+      ////////////////////////////////////////
+      void restart()
+      {
+        param_nlt** p = &avail; // move next to the end of avail
+        while (*p != NULL)
+          p = &((*p)->next);
+        *p = next;
+        this->init();
       }
 
       void check_validity(param_siz& siz);
-      void set_type3_transformation(ui32 comp_num, bool enable);
-      bool get_type3_transformation(ui32 comp_num, ui8& bit_depth, 
-                                    bool& is_signed) const;
+      void set_nonlinear_transform(ui32 comp_num, ui8 nl_type);
+      bool get_nonlinear_transform(ui32 comp_num, ui8& bit_depth,
+                                   bool& is_signed, ui8& nl_type) const;
       bool write(outfile_base* file) const;
       void read(infile_base* file);
 
     private:
-      const param_nlt* get_comp_object(ui32 comp_num) const;
-      param_nlt* get_comp_object(ui32 comp_num);
+      ////////////////////////////////////////
+      void init()
+      {
+        Lnlt = 6;
+        Cnlt = special_comp_num::ALL_COMPS; // default
+        BDnlt = 0;
+        Tnlt = nonlinearity::OJPH_NLT_UNDEFINED;
+        enabled = false; next = NULL;
+      }
+
+      ////////////////////////////////////////
+      void destroy()
+      {
+        if (avail)
+          delete avail;
+        if (next) {
+          delete next;
+          next = NULL;
+        }
+      }
+
+    private:
+      const param_nlt* get_nlt_object(ui32 comp_num) const;
+      param_nlt* get_nlt_object(ui32 comp_num);
       param_nlt* add_object(ui32 comp_num);
       bool is_any_enabled() const;
       void trim_non_existing_components(ui32 num_comps);
@@ -735,13 +845,14 @@ namespace ojph {
       ui8 Tnlt;          // Type of non-linearity
       bool enabled;      // true if this object is used
       param_nlt* next;   // for chaining NLT markers
-      bool alloced_next; // true if next was allocated, not just set to an
-                         // existing object
 
-      // The top level param_nlt object is not allocated, but as part of 
+      // The top level param_nlt object is not allocated, but as part of
       // codestream, and is used to manage allocated next objects.
       // next holds a list of param_nlt objects, which are managed by the top
       // param_nlt object.
+
+    private: // on restart, already allocated param_nlt objs are stored here
+      param_nlt* avail;
     };
 
     ///////////////////////////////////////////////////////////////////////////
@@ -769,15 +880,13 @@ namespace ojph {
           Ccap[0] |= 0x0020;
         Ccap[0] &= 0xFFE0;
         ui32 Bp = 0;
-        ui32 B = qcd.get_MAGBp();
+        ui32 B = qcd.get_MAGB();
         if (B <= 8)
           Bp = 0;
         else if (B < 28)
           Bp = B - 8;
-        else if (B < 48)
-          Bp = 13 + (B >> 2);
         else
-          Bp = 31;
+          Bp = 13 + (B >> 2);
         Ccap[0] = (ui16)(Ccap[0] | (ui16)Bp);
       }
 
@@ -877,10 +986,19 @@ namespace ojph {
       };
 
     public: // member functions
-      param_dfs() { init(); }
-      ~param_dfs() { if (next) delete next; }
-      void init() 
-      { Ldfs = Sdfs = Ids = 0; memset(Ddfs, 0, sizeof(Ddfs)); next = NULL; }
+      param_dfs() { avail = NULL; init(); }
+      ~param_dfs() { destroy(); }
+
+      ////////////////////////////////////////
+      void restart()
+      {
+        param_dfs** p = &avail; // move next to the end of avail
+        while (*p != NULL)
+          p = &((*p)->next);
+        *p = next;
+        this->init();
+      }
+
       bool read(infile_base *file);
       bool exists() const { return Ldfs != 0; }
 
@@ -893,6 +1011,22 @@ namespace ojph {
                            ui32 subband) const;
       point get_res_downsamp(ui32 skipped_resolutions) const;
 
+    private:
+      ////////////////////////////////////////
+      void init()
+      { Ldfs = Sdfs = Ids = 0; memset(Ddfs, 0, sizeof(Ddfs)); next = NULL; }
+
+      ////////////////////////////////////////
+      void destroy()
+      {
+        if (avail)
+          delete avail;
+        if (next) {
+          delete next;
+          next = NULL;
+        }
+      }
+
     private: // member variables
       ui16 Ldfs;       // length of the segment marker
       ui16 Sdfs;       // index of this DFS marker segment
@@ -900,6 +1034,9 @@ namespace ojph {
       ui8 Ddfs[8];     // a string defining number of decomposition sub-levels
                        // 8 bytes should be enough for 32 levels
       param_dfs* next; // used for linking other dfs segments
+
+    private: // on restart, already allocated param_dfs objs are stored here
+      param_dfs* avail;
     };
 
     ///////////////////////////////////////////////////////////////////////////
@@ -934,44 +1071,37 @@ namespace ojph {
     {
       // Limitations:
       // Arbitrary filters (ARB) are not supported
-      // Up to 6 steps are supported -- more than 6 are not supported
       // Only one coefficient per step -- first order filter
       // Only even-indexed subsequence in first reconstruction step,
       //   m_init = 0 is supported
+    private:
+      enum : ui8 { // this is not a real parameter; it used for debugging
+        OJPH_ATK_TOP = 0,
+        OJPH_ATK_NONTOP = 1
+      };
 
     public: // member functions
-      param_atk() { init(); }
-      ~param_atk() {
-        if (next && alloced_next) {
-          delete next;
-          next = NULL;
-        }
-        if (d != NULL && d != d_store) {
-          delete[] d;
-          init(false);
-        }
+      param_atk(param_atk *top_atk = NULL, ui8 type = OJPH_ATK_TOP) {
+        avail = NULL;
+        d = d_store;
+        max_steps = sizeof(d_store) / sizeof(lifting_step);
+        this->type = type;
+        init(this);
       }
+      ~param_atk() { destroy(); }
+
+      ////////////////////////////////////////
+      void restart()
+      {
+        assert(type == OJPH_ATK_TOP);
+        param_atk** p = &avail; // move next to the end of avail
+        while (*p != NULL)
+          p = &((*p)->next);
+        *p = next;
+        this->init(this);
+      }
+
       bool read(infile_base *file);
-      bool read_coefficient(infile_base *file, float &K);
-      bool read_coefficient(infile_base *file, si16 &K);
-      void init(bool clear_all = true) { 
-        if (clear_all) 
-        {
-          Latk = Satk = 0;
-          Katk = 0.0f;
-          Natk = 0;
-          d = NULL;
-          max_steps = 0;
-          memset(d_store, 0, sizeof(d_store));
-          next = NULL;
-          alloced_next = false;
-        }
-        d = d_store; max_steps = sizeof(d_store) / sizeof(lifting_step);
-      }
-      void init_irv97();
-      void init_rev53();
-      void link(param_atk* next) 
-      { assert(this->next == NULL); this->next = next; alloced_next = false; }
 
       ui8 get_index() const { return (ui8)(Satk & 0xFF); }
       int get_coeff_type() const { return (Satk >> 8) & 0x7; }
@@ -979,12 +1109,53 @@ namespace ojph {
       bool is_reversible() const { return (Satk & 0x1000) != 0; }
       bool is_m_init0() const { return (Satk & 0x2000) == 0; }
       bool is_using_ws_extension() const { return (Satk & 0x4000) != 0; }
-      const param_atk* get_atk(int index) const;
-      const lifting_step* get_step(ui32 s) const 
+      param_atk* get_atk(int index);
+      const lifting_step* get_step(ui32 s) const
       { assert(s < Natk); return d + s; }
       ui32 get_num_steps() const { return Natk; }
       float get_K() const { return Katk; }
 
+    private:
+      ////////////////////////////////////////
+      void init(param_atk *top_atk)
+      {
+        Latk = Satk = 0;
+        Katk = 0.0f;
+        Natk = 0;
+        next = NULL;
+        this->top_atk = top_atk;
+        if (d == NULL) {
+          d = d_store;
+          max_steps = sizeof(d_store) / sizeof(lifting_step);
+        }
+        memset(d, 0, max_steps * sizeof(lifting_step));
+      }
+
+      ////////////////////////////////////////
+      void destroy()
+      {
+        assert(type == OJPH_ATK_TOP);
+        if (avail)
+          delete avail;
+        if (next) {
+          delete next;
+          next = NULL;
+        }
+        if (d != NULL && d != d_store) {
+          delete[] d;
+          d = d_store;
+          max_steps = sizeof(d_store) / sizeof(lifting_step);
+        }
+      }
+
+  private:
+      bool read_coefficient(infile_base *file, float &K);
+      bool read_coefficient(infile_base *file, si16 &K);
+
+      void init_irv97();
+      void init_rev53();
+      param_atk* add_object();
+
     private: // member variables
       ui16 Latk;         // structure length
       ui16 Satk;         // carries a variety of information
@@ -995,8 +1166,13 @@ namespace ojph {
       lifting_step d_store[6];   // lifting step coefficient
       param_atk* next;   // used for chaining if more than one atk segment
                          // exist in the codestream
-      bool alloced_next; // true if next was allocated, not just set to an
-                         // existing object
+      ui8 type;          // marker type -- this not a standard type; it is used
+                         // for debugging
+      param_atk* top_atk;// This is the top level atk, from which all atk
+                         // objects are derived
+
+    private: // on restart, already allocated param_atk objs are stored here
+      param_atk* avail;
     };
   } // !local namespace
 } // !ojph namespace
